; Assembler source for FPA support code - simple stand-alone veneer
;
; Copyright (C) Advanced RISC Machines Limited, 1992-7 (now named ARM Limited).
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions are met:
;
;     * Redistributions of source code must retain the above copyright notice,
;       this list of conditions and the following disclaimer.
;     * Redistributions in binary form must reproduce the above copyright
;       notice, this list of conditions and the following disclaimer in the
;       documentation and/or other materials provided with the distribution.
;     * Neither the name of Advanced RISC Machines Limited nor ARM Limited nor
;       the names of its contributors may be used to endorse or promote
;       products derived from this software without specific prior written
;       permission.
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
; AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
; ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
; LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
; SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
; INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
; CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
; ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.

        AREA    Simple_FPASC, CODE

                        GBLA    ExpectsCoreVersion
ExpectsCoreVersion      SETA    1

        MACRO
$label  AdrWS   $reg
$label  ADRL    $reg,Workspace
        MEND

        MACRO
$label  FPInstrDone $opt
        [ "$opt" = ""
$label
        |
$label    TEQ     R9,#0         ;Allow the optimisation, assuming a normal
                                ;  instruction address
        ]
        MEND

        IMPORT  veneer_newhandlers
        IMPORT  veneer_preservehandlers
        IMPORT  veneer_restorehandlers
        IMPORT  FPA_ABSENT
        IMPORT  FPASC_UNHANDLED

; The stack areas.

UndefStack
        %       512
UndefStackEnd

SvcStack
        %       128
SvcStackEnd

; The initialisation routine.

FPASC_STARTUP
        EXPORT  FPASC_STARTUP

; First set up supervisor mode stack and undefined instruction mode stack.

        ADR     R13,SvcStackEnd
        MRS     R0,CPSR
        BIC     R1,R0,#&1F
        ORR     R1,R1,#&1B
        MSR     CPSR_c,R1
        NOP
        ADR     R13,UndefStackEnd
        MSR     CPSR_c,R0
        NOP

; Preserve the return link.

        STMFD   R13!,{R14}

; Get hold of the workspace pointer, in R10.

        AdrWS   R10

; Preserve the current undefined instruction handler.

        BL      veneer_preservehandlers

; Initialise the FPASC core, using our own handler as the "next handler in
; the chain".

        LDR     R1,=FPASC_UNHANDLED
        BL      core_initws

; Check we've got the FPA.

        TEQ     R0,#SysID_FPA
        BNE     FPA_ABSENT

; Activate the one and only floating point context.

        BL      core_activatecontext

; Return.

        LDMFD   R13!,{PC}

        LTORG

        END
